"""
evaluate_balance.py
====================================
Script to evaluate the performance of using class weights.
"""

from tools import data_loader
from utility.config import load_config
import csv
import paths as pt
from pathlib import Path
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import cross_validate
from utility.metrics import compute_mean, compute_std
from sklearn.ensemble import RandomForestClassifier

def make_model(class_weight=None):
    return RandomForestClassifier(n_estimators=800,
                                  class_weight=class_weight,
                                  random_state=0)

def main():
    cases = ["Complete", "Compliance", "Fall", "Risk"]
    for case in cases:
        output_filename = f"{case} model balance.csv"
        header = ['clf', 'version', 'accuracy_mean', 'accuracy_std',
                  'precision_mean', 'precision_std', 'recall_mean',
                  'recall_std', 'roc_auc_mean', 'roc_auc_std',
                  'pr_auc_mean', 'pr_auc_std', 'f1_mean', 'f1_std']
        with open(Path.joinpath(pt.REPORTS_DIR, output_filename), 'w',
                  encoding='UTF8', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(header)

        settings = load_config(pt.CONFIGS_DIR, f"{case.lower()}.yaml")
        if case == "Complete":
            dl = data_loader.CompleteDataLoader(pt.PROCESSED_DATA_DIR,
                                                "complete_emb.csv", settings)
        elif case == "Compliance":
            dl = data_loader.ComplianceDataLoader(pt.PROCESSED_DATA_DIR,
                                                  "compliance_emb.csv", settings)
        elif case == "Fall":
            dl = data_loader.FallDataLoader(pt.PROCESSED_DATA_DIR,
                                            "fall_emb.csv", settings)
        else:
            dl = data_loader.RiskDataLoader(pt.PROCESSED_DATA_DIR,
                                            "risk_emb.csv", settings)
        X, y = dl.load_data().get_data()

        versions = ['NoCW', 'CW']
        metrics = ['accuracy', 'precision', 'recall', 'roc_auc', 'average_precision', 'f1']
        for version in versions:
            if version == "NoCW":
                model = make_model()
                kfold = StratifiedKFold(n_splits=5, shuffle=True, random_state=0)
                results = cross_validate(model, X, y, cv=kfold, scoring=metrics)
            else:
                model = make_model(class_weight="balanced")
                kfold = StratifiedKFold(n_splits=5, shuffle=True, random_state=0)
                results = cross_validate(model, X, y, cv=kfold, scoring=metrics)

            with open(Path.joinpath(pt.REPORTS_DIR, output_filename), 'a',
                      encoding='UTF8', newline='') as f:
                writer = csv.writer(f)
                data = ["RF", version]
                for metric in metrics:
                    mean = compute_mean(results[f'test_{metric}'])
                    std = compute_std(results[f'test_{metric}'])
                    data.extend((mean, std))
                writer.writerow(data)

if __name__ == '__main__':
    main()
