"""
evaluate_classification_cases.py
====================================
Script to make a baseline evaluation of classification cases
using CV with different datasets and classifers.
"""

import numpy as np
import paths as pt
from typing import List
from tools import file_writer, preprocessor, data_loader
from tools.classifiers import KnnClassifier, SvmClassifier, LrClassifier
from tools.classifiers import XgbClassifier, RfClassifier, MlpClassifier
from pathlib import Path
import csv
from utility.config import load_config
from utility.metrics import compute_mean, compute_std
from io import BytesIO
import shutil

NUM_ITER = 1

def get_version_subtitle(version):
    if version == "NoAts":
        return "without Ats and/or Ex columns"
    elif version == "Embedded":
        return "with embeddings"
    else:
        return "with counts"

def prepare_data(X, y, settings):
    features_to_normalize = settings['features_to_normalize']
    features_to_scale = settings['features_to_scale']
    X = preprocessor.normalize_data(X, features_to_normalize)
    X = preprocessor.scale_data(X, features_to_scale)
    X = np.array(X)
    y = np.array(y)
    return X, y

def load_data_embedded(case, settings):
    if case == "Complete":
        dl = data_loader.CompleteDataLoader(pt.PROCESSED_DATA_DIR,
                                            "complete_emb.csv", settings)
    elif case == "Compliance":
        dl = data_loader.ComplianceDataLoader(pt.PROCESSED_DATA_DIR,
                                              "compliance_emb.csv", settings)
    elif case == "Fall":
        dl = data_loader.FallDataLoader(pt.PROCESSED_DATA_DIR,
                                        "fall_emb.csv", settings)
    else:
        dl = data_loader.RiskDataLoader(pt.PROCESSED_DATA_DIR,
                                        "risk_emb.csv", settings)
    X, y = dl.load_data().get_data()
    return X, y

def load_data_count(case, settings):
    if case == "Complete":
        dl = data_loader.CompleteDataLoader(pt.PROCESSED_DATA_DIR,
                                            "complete_count.csv", settings)
    elif case == "Compliance":
        dl = data_loader.ComplianceDataLoader(pt.PROCESSED_DATA_DIR,
                                              "compliance_count.csv", settings)
    elif case == "Fall":
        dl = data_loader.FallDataLoader(pt.PROCESSED_DATA_DIR,
                                        "fall_count.csv", settings)
    else:
        dl = data_loader.RiskDataLoader(pt.PROCESSED_DATA_DIR,
                                        "risk_count.csv", settings)
    X, y = dl.load_data().get_data()
    return X, y

def load_data_ordinal(case, settings):
    if case == "Complete":
        dl = data_loader.CompleteDataLoader(pt.PROCESSED_DATA_DIR,
                                            "complete_ordinal.csv", settings)
    elif case == "Compliance":
        dl = data_loader.ComplianceDataLoader(pt.PROCESSED_DATA_DIR,
                                              "compliance_ordinal.csv", settings)
    elif case == "Fall":
        dl = data_loader.FallDataLoader(pt.PROCESSED_DATA_DIR,
                                        "fall_ordinal.csv", settings)
    else:
        dl = data_loader.RiskDataLoader(pt.PROCESSED_DATA_DIR,
                                        "risk_ordinal.csv", settings)
    X, y = dl.load_data().get_data()
    return X, y

def load_data_ohe(case, settings):
    if case == "Complete":
        dl = data_loader.CompleteDataLoader(pt.PROCESSED_DATA_DIR,
                                            "complete_ohe.csv", settings)
    elif case == "Compliance":
        dl = data_loader.ComplianceDataLoader(pt.PROCESSED_DATA_DIR,
                                              "compliance_ohe.csv", settings)
    elif case == "Fall":
        dl = data_loader.FallDataLoader(pt.PROCESSED_DATA_DIR,
                                        "fall_ohe.csv", settings)
    else:
        dl = data_loader.RiskDataLoader(pt.PROCESSED_DATA_DIR,
                                        "risk_ohe.csv", settings)
    X, y = dl.load_data().get_data()
    return X, y

def main():
    clf_names = ['KNN', 'SVM', 'LR', 'XGB', 'RF', 'MLP']
    num_clfs = len(clf_names)
    metrics = ['accuracy', 'precision', 'recall', 'roc_auc', 'average_precision', 'f1']
    cases = ["Complete"]
    for case in cases:
        target_settings = load_config(pt.CONFIGS_DIR, f'{case.lower()}.yaml')
        output_filename = f"{case} model baseline.csv"
        header = ['clf', 'version', 'accuracy_mean', 'accuracy_std',
                  'precision_mean', 'precision_std', 'recall_mean',
                  'recall_std', 'roc_auc_mean', 'roc_auc_std',
                  'pr_auc_mean', 'pr_auc_std', 'f1_mean', 'f1_std']
        with open(Path.joinpath(pt.REPORTS_DIR, output_filename), 'w',
                  encoding='UTF8', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(header)
        versions = ["Embedded"]
        for version in versions:
            if version == "Embedded":
                X, y = load_data_embedded(case, target_settings)
            elif version == "Counts":
                X, y = load_data_count(case, target_settings)
            elif version == "Ordinal":
                X, y = load_data_ordinal(case, target_settings)
            else:
                X, y = load_data_ohe(case, target_settings)

            X, y = prepare_data(X, y, target_settings)
            results = train_clf(X, y, version, output_filename, metrics, NUM_ITER)
            subtitle = get_version_subtitle(version)
            make_plots(results, metrics, NUM_ITER, num_clfs,
                       clf_names, case, version, subtitle)

def train_clf(X, y, version, output_filename, metrics, num_iter):
    iteration_results = list()
    for k in range(num_iter):
        results = {'KNN': KnnClassifier(X, y).evaluate(metrics, k),
                   'SVM': SvmClassifier(X, y).evaluate(metrics, k),
                   'LR': LrClassifier(X, y).evaluate(metrics, k),
                   'XGB': XgbClassifier(X, y).evaluate(metrics, k),
                   'RF': RfClassifier(X, y).evaluate(metrics, k),
                   'MLP': MlpClassifier(X, y).evaluate(metrics, k)}
        iteration_results.append(results)

    for iter_result in iteration_results:
        for clf_name, result in iter_result.items():
            with open(Path.joinpath(pt.REPORTS_DIR, output_filename), 'a',
                      encoding='UTF8', newline='') as f:
                writer = csv.writer(f)
                data = [clf_name, version]
                for metric in metrics:
                    mean = compute_mean(result[metric])
                    std = compute_std(result[metric])
                    data.extend((mean, std))
                writer.writerow(data)

    return iteration_results

def make_plots(results: np.ndarray, metrics: List[str], num_iter: int,
               num_clfs: int, clf_names, case: str, version: str, subtitle):
    for metric in metrics:
        total_means, total_stds = list(), list()
        for iter_result in results:
            means, stds = np.zeros(num_clfs), np.zeros(num_clfs)
            for i, (_, result) in enumerate(iter_result.items()):
                means[i] = compute_mean(result[metric])
                stds[i] = compute_std(result[metric])
            total_means.append(means)
            total_stds.append(stds)
        total_means = np.stack(total_means, axis=-1)
        total_stds = np.stack(total_stds, axis=-1)
        outfile = BytesIO()
        file_path = pt.REPORTS_PLOTS_DIR
        file_name = f"{case} version {version} - {metric}.pdf"
        with open(Path.joinpath(file_path, file_name), 'wb') as fd:
            file_writer.write_cv_plot(total_means, total_stds,
                                      metric, num_iter, clf_names,
                                      file_name, subtitle, outfile)
            outfile.seek(0)
            shutil.copyfileobj(outfile, fd)

if __name__ == '__main__':
    main()
