"""
evaluate_survival_case.py
====================================
Script to make a baseline evaluation of the survival case using CV.
"""

import paths as pt
from tools import data_loader
from utility.config import load_config
from utility.data import write_csv
from sklearn.model_selection import KFold
from sksurv.ensemble import GradientBoostingSurvivalAnalysis
from sksurv.metrics import (concordance_index_censored,
                            concordance_index_ipcw,
                            integrated_brier_score)
import pandas as pd
import numpy as np

def main():
    target_settings = load_config(pt.CONFIGS_DIR, "alarm.yaml")
    dl = data_loader.AlarmDataLoader(pt.PROCESSED_DATA_DIR,
                                     "alarm_emb.csv",
                                     target_settings).load_data()
    X, y = dl.get_data()

    model = GradientBoostingSurvivalAnalysis(n_estimators=100,
                                             learning_rate=0.1,
                                             max_depth=14,
                                             loss='coxph',
                                             min_samples_split=4,
                                             max_features='log2',
                                             subsample=0.4,
                                             dropout_rate=0.7,
                                             random_state=0)

    model.fit(X, y)

    kf = KFold(n_splits=5, shuffle=True, random_state=0)
    c_index_harells = list()
    c_index_unos = list()
    brier_scores = list()
    for train, test in kf.split(X, y):
        model.fit(X.iloc[train], y[train])
        prediction = model.predict(X.iloc[test])

        c_harrell = concordance_index_censored(y[test]["Status"],
                                               y[test]["Days_to_alarm"],
                                               prediction)
        c_uno = concordance_index_ipcw(y[train], y[test], prediction)
        lower, upper = np.percentile(y["Days_to_alarm"], [10, 90])
        alarm_times = np.arange(lower, upper+1)
        surv_prob = np.row_stack([fn(alarm_times)
                                  for fn in model.predict_survival_function(X.iloc[test])])
        brier_score = integrated_brier_score(y[train], y[test],
                                             surv_prob, alarm_times)

        c_index_harells.append(c_harrell[0])
        c_index_unos.append(c_uno[0])
        brier_scores.append(brier_score)

    c_index_harell_mean = np.mean(c_index_harells)
    c_index_uno_mean = np.mean(c_index_unos)
    brier_score_mean = np.mean(brier_scores)

    df = pd.DataFrame([['GradientBoosting'] + [c_index_harell_mean,
                                               c_index_uno_mean,
                                               brier_score_mean]],
                      columns=['Method', 'c_harrell', 'c_uno', 'brier_score'])
    write_csv(df, pt.REPORTS_DIR, 'Alarm model baseline.csv')

if __name__ == '__main__':
    main()
