"""
find_best_ats_resolution.py
====================================
Script to find the best ATS resolution for Complete case.
"""

import paths as pt
from pathlib import Path
from utility.config import load_config
from tools import labeler, preprocessor
from utility.data import read_csv
from utility.embedder import encode_dataframe
from sklearn.model_selection import train_test_split
import xgboost as xgb
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

def main():
    step_size = 10
    for idx in range(1, 10+1, step_size):
        logloss_train, logloss_test = list(), list()
        auc_train, auc_test = list(), list()

        converters = {'CitizenId': str}
        file_path = pt.INTERIM_DATA_DIR
        file_name = 'screenings.csv'
        screenings = read_csv(file_path, file_name, converters=converters)

        df = screenings.copy()
        data_settings = load_config(pt.CONFIGS_DIR, 'data.yaml')
        accum_screenings = labeler.accumulate_screenings(df, data_settings)

        target_settings = load_config(pt.CONFIGS_DIR, 'complete.yaml')
        features = target_settings['features']

        for ats_res in range(idx, idx+step_size):
            df = labeler.make_complete_label(accum_screenings)
            df = preprocessor.split_cat_columns(df, col_to_split='Ats',
                                                tag='Ats', resolution=ats_res)
            ats_cols = df.filter(regex='Ats', axis=1)
            df = pd.concat([df[features], ats_cols, df[["Complete"]]], axis=1)

            emb_cols = df.filter(regex='((\d+)[Ats])\w+', axis=1)
            n_numerical_cols = df.shape[1] - emb_cols.shape[1] - 1
            df_to_enc = df.iloc[:,n_numerical_cols:]
            ats_cols = [str(i)+'Ats' for i in range(1, ats_res+1)]
            df = df.drop(ats_cols, axis=1)

            df_enc = encode_dataframe(df=df_to_enc,
                                      target_name=target_settings['target_name'],
                                      metrics=target_settings['metrics'],
                                      batch_size=target_settings['batch_size'],
                                      train_ratio=target_settings['train_ratio'],
                                      epochs=target_settings['num_epochs'],
                                      optimizer=target_settings['optimizer'],
                                      network_layers=target_settings['network_layers'],
                                      verbose=target_settings['verbose'],
                                      model_path=None,
                                      enable_emb_viz=False)

            df = pd.concat([df.drop("Complete", axis=1), df_enc,
                            df.pop("Complete")], axis=1)

            X = df.drop("Complete", axis=1)
            y = df.pop("Complete")

            X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3,
                                                                stratify=y,
                                                                random_state=0)

            params = {"n_estimators": 200,
                      "objective": "binary:logistic",
                      "use_label_encoder": False,
                      "learning_rate": 0.1,
                      "random_state": 0}
            model = xgb.XGBClassifier(**params)

            eval_set = [(X_train, y_train), (X_test, y_test)]
            eval_metric = ["auc", "logloss"]
            model.fit(X_train, y_train,
                      eval_set=eval_set,
                      eval_metric=eval_metric,
                      verbose=False)

            results = model.evals_result()
            logloss_train.append((ats_res, results['validation_0']['logloss']))
            logloss_test.append((ats_res, results['validation_1']['logloss']))
            auc_train.append((ats_res, results['validation_0']['auc']))
            auc_test.append((ats_res, results['validation_1']['auc']))

        _, ax = plt.subplots()
        x_axis = range(0, 200)
        cmap = plt.cm.coolwarm
        for entry in logloss_train:
            ax.plot(x_axis, entry[1], label=f'Train ATS={entry[0]}', color=cmap(0.))
            ax.text(x_axis[-1], entry[1][-1], f'ATS={entry[0]}', fontsize=2)
        for entry in logloss_test:
            ax.plot(x_axis, entry[1], label=f'Test ATS={entry[0]}', color=cmap(1.))
            ax.text(x_axis[-1], entry[1][-1], f'ATS={entry[0]}', fontsize=2)

        custom_lines = [Line2D([0], [0], color=cmap(0.), lw=4),
                        Line2D([0], [0], color=cmap(1.), lw=4)]

        ax.legend(custom_lines, ['Train logloss',
                                 'Test logloss'])
        file_name = f"XGBoost logloss for ATS {idx}-{idx+(step_size-1)}"
        plt.ylabel('Logloss')
        plt.xlabel('Iterations')
        plt.title(file_name)
        plt.savefig(Path.joinpath(pt.REPORTS_PLOTS_DIR, f'{file_name}.pdf'),
                    dpi=300,
                    bbox_inches = "tight")

if __name__ == "__main__":
    main()
