"""
find_best_features.py
====================================
Script to find best features for cases using CV.
"""

import numpy as np
import paths as pt
from tools import file_writer, data_loader
from sklearn.model_selection import StratifiedKFold
import xgboost as xgb
from sklearn.metrics import accuracy_score
import pandas as pd
from utility.metrics import gini_xgb
import shap
from typing import List
from utility.config import load_config
from io import BytesIO, StringIO
import shutil
from pathlib import Path

NUM_ITERATIONS = 5

def main():
    cases = ["Complete", "Compliance", "Fall", "Risk"]
    for case in cases:
        settings = load_config(pt.CONFIGS_DIR, f"{case.lower()}.yaml")
        if case == "Complete":
            dl = data_loader.CompleteDataLoader(pt.PROCESSED_DATA_DIR,
                                                "complete_emb.csv", settings)
        elif case == "Compliance":
            dl = data_loader.ComplianceDataLoader(pt.PROCESSED_DATA_DIR,
                                                  "compliance_emb.csv", settings)
        elif case == "Fall":
            dl = data_loader.FallDataLoader(pt.PROCESSED_DATA_DIR,
                                            "fall_emb.csv", settings)
        else:
            dl = data_loader.RiskDataLoader(pt.PROCESSED_DATA_DIR,
                                            "risk_emb.csv", settings)

        X, y = dl.load_data().get_data()
        features = dl.get_features()

        X = np.array(X)
        y = np.array(y)

        total_shap_df  = pd.DataFrame()
        for seed in range(NUM_ITERATIONS):
            print('#'*40, '{} of {} iterations'.format(seed+1, NUM_ITERATIONS), '#' * 40)
            temp_shap_df = get_best_shap_features(X, y, features, seed)
            total_shap_df = pd.concat([total_shap_df, temp_shap_df])

        shap_sorted_df = total_shap_df.groupby('feature').mean().sort_values('shap_values',
                                                                            ascending=False)
        shap_sorted_df = shap_sorted_df.reset_index()

        importances = list(shap_sorted_df['shap_values'])
        features = list(shap_sorted_df['feature'])

        outfile = BytesIO()
        file_path = pt.REPORTS_PLOTS_DIR
        file_name = f"{case} SHAP feature values.pdf"
        with open(Path.joinpath(file_path, file_name), 'wb') as fd:
            file_writer.write_shap_importance_plot(features,
                                                   importances,
                                                   file_name,
                                                   outfile)
            outfile.seek(0)
            shutil.copyfileobj(outfile, fd)

        outfile = StringIO()
        file_path = pt.REPORTS_PLOTS_DIR
        file_name = f"{case} model features.csv"
        with open(Path.joinpath(file_path, file_name), 'w', newline='') as fd:
            file_writer.write_csv(shap_sorted_df, outfile)
            outfile.seek(0)
            shutil.copyfileobj(outfile, fd)

def get_best_shap_features(X: np.ndarray, y: np.ndarray,
                           cols: List[str], seed: int):
    K = 5
    kf = StratifiedKFold(n_splits=K, random_state=seed, shuffle=True)
    feat_importance_df  = pd.DataFrame()

    neg, pos = np.bincount(y)
    scale_pos_weight = neg / pos

    model = xgb.XGBClassifier(n_estimators=400,
                              learning_rate=0.1,
                              objective='binary:logistic',
                              scale_pos_weight=scale_pos_weight,
                              eval_metric='logloss',
                              use_label_encoder=False,
                              random_state=seed)

    acc_score_list = list()
    for i, (train_index, test_index) in enumerate(kf.split(X, y)):
        print('='*30, '{} of {} folds'.format(i+1, K), '='*30)
        X_train, X_val = X[train_index], X[test_index]
        y_train, y_val = y[train_index], y[test_index]
        model.fit(X_train, y_train,
                  eval_set=[(X_train, y_train), (X_val, y_val)],
                  eval_metric=gini_xgb,
                  verbose=0)
        y_val_pred = model.predict_proba(X_val)[:,1]
        y_scores_new = (y_val_pred > 0.5)
        acc_score_list.append(np.around(accuracy_score(y_val, y_scores_new), decimals=3))
        shap_values = shap.TreeExplainer(model).shap_values(X_train)
        fold_importance_df  = pd.DataFrame()
        fold_importance_df['feature'] = cols
        fold_importance_df['shap_values'] = np.around(abs(np.array(shap_values)[:,:]).mean(0), decimals=2)
        fold_importance_df['feat_imp'] = np.around(model.feature_importances_, decimals=2)
        feat_importance_df = pd.concat([feat_importance_df, fold_importance_df])

    mean_acc_score = np.mean(acc_score_list)
    print(f"Mean accuracy: {mean_acc_score}")
    feat_importance_df_shap = feat_importance_df.groupby('feature').mean().sort_values('shap_values',
                                                                                       ascending=False)
    feat_importance_df_shap = feat_importance_df_shap.reset_index()
    return feat_importance_df_shap

if __name__ == '__main__':
    main()
