"""
main.py
====================================
The main FastAPI module class.
"""

import uvicorn
import pydantic
import os
import csv
import joblib
import yaml
import pandas as pd
import numpy as np
from typing import Optional
from fastapi import Depends, FastAPI, HTTPException, Request
from fastapi.responses import JSONResponse
from fastapi_jwt_auth import AuthJWT
from fastapi_jwt_auth.exceptions import AuthJWTException
from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials
from sksurv.ensemble import GradientBoostingSurvivalAnalysis
from xgboost import XGBClassifier
from decimal import Decimal

app = FastAPI(title='AIR API', version='1.0',
              description='An API that classifies citizens based on data')

class User(pydantic.BaseModel):
    username: str
    password: str

class Settings(pydantic.BaseModel):
    authjwt_secret_key: str = "secret"

class JWTBearer(HTTPBearer):
    def __init__(self, auto_error: bool = True):
        super(JWTBearer, self).__init__(auto_error=auto_error)

    async def __call__(self, request: Request):
        credentials: HTTPAuthorizationCredentials = await super(JWTBearer, self).__call__(request)
        if credentials:
            if not credentials.scheme == "Bearer":
                raise HTTPException(status_code=403, detail="Invalid authentication scheme.")
            if not self.verify_jwt(request, credentials.credentials):
                raise HTTPException(status_code=403, detail="Invalid token or expired token.")
            return credentials.credentials
        else:
            raise HTTPException(status_code=403, detail="Invalid authorization code.")

    # noinspection PyBroadException
    @staticmethod
    def verify_jwt(request: Request, token: Optional[str]) -> bool:
        is_token_valid: bool = False
        auth_jwt = AuthJWT(request, token)
        try:
            payload = auth_jwt.get_raw_jwt()
        except Exception:
            payload = None
        if payload:
            is_token_valid = True
        return is_token_valid

@AuthJWT.load_config
def get_config():
    return Settings()

@app.exception_handler(AuthJWTException)
def authjwt_exception_handler(request: Request, exc: AuthJWTException):
    return JSONResponse(
        status_code=exc.status_code,
        content={"detail": exc.message}
    )

class InputData(pydantic.BaseModel):
    Gender: int
    BirthYear: int
    LoanPeriod: float
    Ats: str

class AlarmOutputData(pydantic.BaseModel):
    EventTimes: list
    SurvivalProbs: list
    AlarmArguments: list

class TrainingOutputData(pydantic.BaseModel):
    CompleteProb: float
    Compliance: int
    CompleteArguments: list

@app.get('/')
def index():
    return {'message': 'AIR API v. 0.1'}

@app.get('/user', dependencies=[Depends(JWTBearer())], tags=["login"])
def user(Authorize: AuthJWT = Depends()):
    Authorize.jwt_required()
    current_user = Authorize.get_jwt_subject()
    return {"user": current_user}

@app.get('/ats', tags=["data"])
def get_all_ats():
    return read_csv('ats.csv')

@app.get("/ats/id", tags=["data"])
def get_ats_by_id(ats_id: str):
    ats = read_csv('ats.csv')
    if ats_id in ats:
        return {"ats_id": ats_id, "ats_name": ats[ats_id]}
    raise HTTPException(status_code=404, detail="Not found")

@app.post('/login', tags=["login"])
def login(user: User, Authorize: AuthJWT = Depends()):
    if user.username != "test" or user.password != "test":
        raise HTTPException(status_code=401, detail="Bad username or password")
    access_token = Authorize.create_access_token(subject=user.username)
    refresh_token = Authorize.create_refresh_token(subject=user.username)
    return {"access_token": access_token, "refresh_token": refresh_token}

@app.post('/refresh', dependencies=[Depends(JWTBearer())], tags=["login"])
def refresh(Authorize: AuthJWT = Depends()):
    Authorize.jwt_refresh_token_required()
    current_user = Authorize.get_jwt_subject()
    new_access_token = Authorize.create_access_token(subject=current_user)
    return {"access_token": new_access_token}

@app.post('/predict_alarm', response_model=AlarmOutputData, tags=["ai"])
def predict_alarm(incoming_data: InputData):
    settings = load_settings('config.yaml')
    ats_resolution = settings['ats_resolution']
    data = validate_data(incoming_data)

    alarm_ats = settings['alarm_ats']
    incoming_ats = [x.strip(' ') for x in data['Ats'].split(",")]
    if any(x in list([alarm_ats]) for x in incoming_ats) == True:
        raise HTTPException(status_code=400, detail=f"Ats {alarm_ats} cannot be in feature set")

    df = prepare_data(data, ats_resolution)
    alarm_model = read_grad_boost_model("alarm_gradboost.joblib")

    df_for_alarm = add_embedding(df.copy(), 'alarm', ats_resolution)
    surv_func = alarm_model.predict_survival_function(df_for_alarm)

    event_times = list(range(1, 13))
    surv_probs = np.array([1 - float(x) for x in surv_func[0].y][:360])
    surv_probs = np.mean(surv_probs.reshape(-1, 30), axis=1)
    surv_probs = list(map((lambda x: float(round(100*Decimal(x), 1))), surv_probs))

    prob_after_one_year = surv_probs[-1]
    alarm_arguments = generate_alarm_arguments(df, ats_resolution, prob_after_one_year)

    return {
        'EventTimes': event_times,
        'SurvivalProbs': surv_probs,
        'AlarmArguments': alarm_arguments
    }

@app.post('/predict_training', response_model=TrainingOutputData, tags=["ai"])
def predict_training(incoming_data: InputData):
    settings = load_settings('config.yaml')
    ats_resolution = settings['ats_resolution']

    data = validate_data(incoming_data)
    df = prepare_data(data, ats_resolution)

    complete_model = read_xgb_model('complete_xgboost.joblib')
    compliance_model = read_xgb_model('compliance_xgboost.joblib')

    df_for_complete = add_embedding(df.copy(), 'complete', ats_resolution)
    df_for_compliance = add_embedding(df.copy(), 'compliance', ats_resolution)

    complete_prob = complete_model.predict_proba(df_for_complete).flatten()[1]
    if complete_prob > 0.5:
        compliance_prob = compliance_model.predict_proba(df_for_compliance).flatten()[1]
    else:
        compliance_prob = 0
    compliance = 0 if compliance_prob < 0.5 else 1
    complete_arguments = generate_complete_arguments(df, ats_resolution, complete_prob)

    return {
        'CompleteProb': float(complete_prob),
        'Compliance': int(compliance),
        'CompleteArguments': complete_arguments
    }

def validate_data(incoming_data: InputData):
    data = incoming_data.dict()
    if not data['Gender'] in [0, 1]:
        raise HTTPException(status_code=400, detail="Invalid gender, check input")
    if not data['BirthYear'] in range(1900, 1999):
        raise HTTPException(status_code=400, detail="Invalid birth year, check input")
    if not data['LoanPeriod'] >= 0:
        raise HTTPException(status_code=400, detail="Invalid loan period, check input")

    if data['Ats'] != '':
        incoming_ats = [x.strip(' ') for x in data['Ats'].split(",")]
        ats_converter = {'ats_id': str}
        ats_df = read_dataframe('ats.csv', converters=ats_converter)
        if all(x in list(ats_df['ats_id']) for x in incoming_ats) != True:
            raise HTTPException(status_code=400, detail="Ats not found, check ats list")
    else:
        data['Ats'] = '0'

    return data

def add_embedding(df: pd.DataFrame, case: str, ats_resolution: int) -> pd.DataFrame:
    for i in range(1, ats_resolution+1):
        embedding = read_embedding(f'{case}_{i}Ats.csv')
        column = f'{i}Ats'
        df[column] = df[column].replace(to_replace=embedding)
        df[column] = pd.to_numeric(df[column])
    return df

def generate_complete_arguments(df: pd.DataFrame, ats_resolution: int, prob: float):
    arguments = list()
    gender_argument = "Kvinder" if int(df.iloc[0].Gender) == 0 else "Mænd"
    arguments.append(gender_argument)

    age_argument = f"på {121 - int(df.iloc[0].BirthYear)} år"
    arguments.append(age_argument)

    number_ats = int(df.iloc[0].NumberAts)
    if int(number_ats) < 1:
        arguments.append("uden hjælpemidler i eget hjem")
    else:
        arguments.append("med følgende hjælpemidler i eget hjem:")

        for i in range(1, ats_resolution+1):
            ats_name = get_ats_name_from_hmi(df.iloc[0][f'{i}Ats'])
            if ats_name != "":
                arguments.append(f'Et {i}. hjælpemiddel af typen {ats_name}')
            else:
                arguments.append(f'Uden et {i}. hjælpemiddel.')

        loan_period = int(df.iloc[0].LoanPeriod)
        loan_period_argument = f"og en gennemsnitlig låneperiode på {loan_period} dage"
        arguments.append(loan_period_argument)

    arguments.append("gennemfører")
    arguments.append(f"med {round(prob*100, 1)}% sandsynlighed")
    return arguments

def generate_alarm_arguments(df: pd.DataFrame,
                             ats_resolution: int,
                             prob_after_one_year: float):
    arguments = list()
    gender_argument = "Kvinder" if int(df.iloc[0].Gender) == 0 else "Mænd"
    arguments.append(gender_argument)

    age_argument = f"på {121 - int(df.iloc[0].BirthYear)} år"
    arguments.append(age_argument)

    number_ats = int(df.iloc[0].NumberAts)
    if int(number_ats) < 1:
        arguments.append("uden hjælpemidler i eget hjem")
    else:
        arguments.append("med følgende hjælpemidler i eget hjem:")

        for i in range(1, ats_resolution+1):
            ats_name = get_ats_name_from_hmi(df.iloc[0][f'{i}Ats'])
            if ats_name != "":
                arguments.append(f'Et {i}. hjælpemiddel af typen {ats_name}')
            else:
                arguments.append(f'Uden et {i}. hjælpemiddel.')

        loan_period = int(df.iloc[0].LoanPeriod)
        loan_period_argument = f"og en gennemsnitlig låneperiode på {loan_period} dage"
        arguments.append(loan_period_argument)

    arguments.append("får efter et år en nødalarm")
    arguments.append(f"med {prob_after_one_year}% sandsynlighed")
    return arguments

def load_settings(file_name):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    with open(f'{dir_path}/{file_name}', 'r') as stream:
        settings = yaml.safe_load(stream)
    return settings

def get_ats_name_from_hmi(ats_id: str):
    ats = read_csv('ats.csv')
    if ats_id in ats:
        return ats[ats_id]
    return ""

def prepare_data(data: dict, ats_resolution: int) -> pd.DataFrame:
    new_data = {k: [v] for k, v in data.items()}
    new_data_df = pd.DataFrame.from_dict(new_data)

    if new_data_df['Ats'][0] == '0':
        new_data_df['NumberAts'] = 0
    else:
        ats = list(filter(None, new_data_df['Ats'][0].split(",")))
        new_data_df['NumberAts'] = len(ats)

    df = split_categorical_columns(new_data_df, col='Ats', tag='Ats',
                                   resolution=ats_resolution)

    df['BirthYear'] = df['BirthYear'].apply(lambda x: int(str(x)[2:]))
    df['LoanPeriod'] = df['LoanPeriod'].astype(int)

    cols_ats = [str(i)+'Ats' for i in range(1, ats_resolution+1)]
    header_list = ['Gender', 'BirthYear', 'LoanPeriod', 'NumberAts'] + cols_ats
    df = df.reindex(columns=header_list)
    df = df.fillna('0')

    return df

def split_categorical_columns(df: pd.DataFrame, col: str,
                              tag: str, resolution: int) -> pd.DataFrame:
    split = pd.DataFrame(df[col].str.split(pat=",", expand=True))
    split = split.drop(split.iloc[:, resolution:], axis=1)
    split = split.fillna(0)
    df = pd.concat([df, split], axis=1)
    df = df.drop(col, axis=1)
    for i in range(0, resolution):
        try:
            df = df.rename(columns={i: f'{i+1}{tag}'})
        except:
            pass
    return df

def read_embedding(filename: str):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    with open(f'{dir_path}/embeddings/{filename}', 'r') as f:
        reader = csv.reader(f)
        embedding_dict = {rows[0]:rows[1] for rows in reader}
    return embedding_dict

def read_grad_boost_model(filename: str) -> GradientBoostingSurvivalAnalysis:
    dir_path = os.path.dirname(os.path.realpath(__file__))
    return joblib.load(f'{dir_path}/models/{filename}')

def read_xgb_model(filename: str) -> XGBClassifier:
    dir_path = os.path.dirname(os.path.realpath(__file__))
    return joblib.load(f'{dir_path}/models/{filename}')

def read_csv(filename:str) -> any:
    dir_path = os.path.dirname(os.path.realpath(__file__))
    with open(f'{dir_path}/data/{filename}', encoding='utf-8') as f:
        reader = csv.reader(f)
        csv_dict = {rows[0]:rows[1] for rows in reader}
    return csv_dict

def read_dataframe(filename:str, converters=None) -> pd.DataFrame:
    dir_path = os.path.dirname(os.path.realpath(__file__))
    df = pd.read_csv(f'{dir_path}/data/{filename}',
                     converters=converters)
    return df

if __name__ == "__main__":
    uvicorn.run(app, port=8000, host="0.0.0.0")
