"""
make_dataset_emb.py
====================================
Script to make a dataset for Complete, Compliance, Fall, Risk and Alarm
cases where categorial features are encoded as entity embeddings.
Takes in a full dataset generated with "make_dataset_full.py" for the
Complete, Compliance, Fall or Risk case, or "make_survival_data.py"
for the Alarm case.
"""

from utility.embedder import encode_dataframe
from utility.config import load_config
from utility.data import read_csv, write_csv
from pathlib import Path
import pandas as pd
import paths as pt

USE_CROSS_VALID = False
USE_GROUPING = False
ENABLE_EMB_VIZ = False

def main():
    for label_name in ["Complete", "Compliance", "Fall", "Risk", "Alarm"]:
        data_settings = load_config(pt.CONFIGS_DIR, 'data.yaml')
        target_settings = load_config(pt.CONFIGS_DIR, f'{label_name.lower()}.yaml')
        ats_resolution = data_settings['ats_resolution']

        if label_name == "Risk":
            ex_resolution = target_settings['ex_resolution']

        if label_name in ["Complete", "Compliance", "Fall", "Alarm"]:
            converters = {str(i)+'Ats':str for i in range(1, ats_resolution+1)}
            file_path = pt.PROCESSED_DATA_DIR
            file_name = f'{label_name.lower()}.csv'
            df = read_csv(file_path, file_name, converters=converters)
        else:
            ex = {str(i)+'Ex':str for i in range(1, ex_resolution+1)}
            ats = {str(i)+'Ats':str for i in range(1, ats_resolution+1)}
            converters = {**ex, **ats}
            file_path = pt.PROCESSED_DATA_DIR
            file_name = f'{label_name.lower()}.csv'
            df = read_csv(file_path, file_name, converters=converters)

        if label_name in ["Complete", "Compliance", "Fall"]:
            emb_cols = df.filter(regex='((\d+)[Ats])\w+', axis=1)
            n_numerical_cols = df.shape[1] - emb_cols.shape[1] - 1
            df_to_enc = df.iloc[:,n_numerical_cols:]
            ats_cols = [str(i)+'Ats' for i in range(1, ats_resolution+1)]
            df = df.drop(ats_cols, axis=1)
        elif label_name == "Alarm":
            emb_cols = df.filter(regex='((\d+)[Ats])\w+', axis=1)
            n_numerical_cols = df.shape[1] - emb_cols.shape[1] - 2
            df_to_enc = df.iloc[:,n_numerical_cols:-1]
            df_to_enc = df_to_enc.rename(columns={'Status': 'Alarm'})
            df_to_enc['Alarm'] = df_to_enc['Alarm'].astype(int)
            ats_cols = [str(i)+'Ats' for i in range(1, ats_resolution+1)]
            df = df.drop(ats_cols, axis=1)
        else:
            ats_cols = [str(i)+'Ats' for i in range(1, ats_resolution+1)]
            ex_cols = [str(i)+'Ex' for i in range(1, ex_resolution+1)]
            df_ats_to_enc = df.filter(regex='Risk|((\d+)[Ats])\w+', axis=1)
            df_ex_to_enc = df.filter(regex='Risk|((\d+)[Ex])\w+', axis=1)
            df = df.drop(ats_cols + ex_cols, axis=1)

        model_path = Path.joinpath(pt.ROOT_DIR, target_settings['model_path'])
        if label_name in ["Complete", "Compliance", "Fall", "Alarm"]:
            df_enc = encode_dataframe(df=df_to_enc,
                                      target_name=target_settings['target_name'],
                                      metrics=target_settings['metrics'],
                                      batch_size=target_settings['batch_size'],
                                      train_ratio=target_settings['train_ratio'],
                                      epochs=target_settings['num_epochs'],
                                      optimizer=target_settings['optimizer'],
                                      network_layers=target_settings['network_layers'],
                                      verbose=target_settings['verbose'],
                                      model_path=model_path,
                                      enable_emb_viz=ENABLE_EMB_VIZ)
        else:
            ats_enc = encode_dataframe(df=df_ats_to_enc,
                                       target_name=target_settings['target_name'],
                                       metrics=target_settings['metrics'],
                                       batch_size=target_settings['batch_size'],
                                       train_ratio=target_settings['train_ratio'],
                                       epochs=target_settings['num_epochs_ats'],
                                       optimizer=target_settings['optimizer'],
                                       network_layers=target_settings['network_layers'],
                                       verbose=target_settings['verbose'],
                                       model_path=model_path,
                                       enable_emb_viz=ENABLE_EMB_VIZ)
            ex_enc = encode_dataframe(df=df_ex_to_enc,
                                      target_name=target_settings['target_name'],
                                      metrics=target_settings['metrics'],
                                      batch_size=target_settings['batch_size'],
                                      train_ratio=target_settings['train_ratio'],
                                      epochs=target_settings['num_epochs_ex'],
                                      optimizer=target_settings['optimizer'],
                                      network_layers=target_settings['network_layers'],
                                      verbose=target_settings['verbose'],
                                      model_path=model_path,
                                      enable_emb_viz=ENABLE_EMB_VIZ)

        if label_name in ["Complete", "Compliance", "Fall"]:
            df = pd.concat([df.drop(label_name, axis=1), df_enc,
                            df.pop(label_name)], axis=1)
        elif label_name == "Alarm":
            df = pd.concat([df.iloc[:,:n_numerical_cols], df_enc,
                            df.iloc[:,-2:]], axis=1)
        else:
            df = pd.concat([df.drop(label_name, axis=1), ats_enc,
                            ex_enc, df.pop(label_name)], axis=1)

        file_path = pt.PROCESSED_DATA_DIR
        file_name = f'{label_name.lower()}_emb.csv'
        write_csv(df, file_path, file_name)

if __name__ == "__main__":
    main()
