"""
make_screening_data.py
====================================
Script to make dataset of screenings to be used for cases.
"""

import numpy as np
import pandas as pd
import paths as pt
from tools import inputter
from utility.config import load_config
from utility.data import read_pickle, write_csv, Data
from pandas.tseries.offsets import DateOffset
from typing import List, Tuple

def main():
    sc = read_pickle(pt.INTERIM_DATA_DIR, 'sc.pkl')
    ss = read_pickle(pt.INTERIM_DATA_DIR, 'ss.pkl')
    td = read_pickle(pt.INTERIM_DATA_DIR, 'td.pkl')
    tc = read_pickle(pt.INTERIM_DATA_DIR, 'tc.pkl')
    ats = read_pickle(pt.INTERIM_DATA_DIR, 'ats.pkl')

    settings = load_config(pt.CONFIGS_DIR, "data.yaml")
    data = Data(sc, ss, td, tc, ats)
    screenings = get_screenings(data, settings)

    file_path = pt.INTERIM_DATA_DIR
    file_name = 'screenings.csv'
    write_csv(screenings, file_path, file_name)

def get_screenings(data, settings):
    ids = create_union_of_ids(data.sc, data.ss, data.td, data.tc)
    all_screenings = pd.DataFrame()

    for id in ids:
        screenings = get_screenings_by_id(data, id, settings)
        all_screenings = pd.concat([all_screenings, screenings], axis=0, ignore_index=True)

    return all_screenings

def get_screenings_by_id(data, id, settings):
    citizen_data = inputter.get_citizen_data(data, id)

    if citizen_data.sc.empty:
        return

    # Record split dates
    first_screening_date = citizen_data.sc.ScreeningDate.iloc[0]
    split_dates = []
    for i, date in enumerate(list(citizen_data.sc.ScreeningDate)):
        if date > (first_screening_date + DateOffset(weeks=15)):
            split_dates.append(citizen_data.sc.ScreeningDate.iloc[i-1])
            first_screening_date = date

    # Split screenings
    split_screenings = []
    split_dates.append(citizen_data.sc.ScreeningDate.iloc[-1]) #add boundary
    for split_date in split_dates:
        df = citizen_data.sc.loc[citizen_data.sc['ScreeningDate'] <= split_date]
        split_screenings.append(df)
        citizen_data.sc = citizen_data.sc.loc[citizen_data.sc['ScreeningDate'] > split_date]

    screenings = pd.DataFrame()
    for n_split, screening_split in enumerate(split_screenings):
        for n_screening, screening in enumerate(screening_split.itertuples(index=False)):
            single_screening = pd.DataFrame([{'CitizenId': id}])
            single_screening['Gender'] = inputter.get_gender(screening_split)
            single_screening['BirthYear'] = inputter.get_birth_year(screening_split)
            single_screening['NumberSplit'] = n_split
            single_screening['NumberScreening'] = n_screening

            # Get time dependent window data
            if n_screening == 0:
                start_date = screening_split.iloc[n_screening].ScreeningDate
            else:
                start_date = screening_split.iloc[n_screening-1].ScreeningDate
            end_date = screening.ScreeningDate
            tds, tcs, _ = inputter.get_screening_data(citizen_data.td, citizen_data.tc,
                                                      citizen_data.ss, start_date, end_date)

            # Record number of trainings after screening end
            n_weeks_after = 0
            n_training_after = 0
            if (n_screening == len(screening_split) - 1):
                screening_date = screening.ScreeningDate
                rating_date_cond = citizen_data.td['RatingDate'] > screening_date
                trainings_after = citizen_data.td.loc[rating_date_cond]
                trainings_end_date = screening_date + DateOffset(weeks=15)
                training_after_cond = (trainings_after['RatingDate'] <= trainings_end_date)
                trainings_before_15 = trainings_after.loc[training_after_cond]
                if len(trainings_before_15) > 0:
                    day_start = trainings_before_15.RatingDate.iloc[0]
                    day_end = trainings_before_15.RatingDate.iloc[-1]
                    n_weeks_after = (day_end - day_start).days / 7
                    n_training_after = len(trainings_before_15)

            # Overwrite start date with rating date if it exists
            start_date = tds.RatingDate.iloc[0] if not tds.empty else start_date

            date_format = '%d-%m-%Y'
            start_date = inputter.convert_date_to_datetime(start_date, date_format)
            end_date = inputter.convert_date_to_datetime(end_date, date_format)
            single_screening['StartDate'] = start_date
            single_screening['EndDate'] = end_date

            n_weeks = inputter.get_interval_length(start_date, end_date)
            single_screening['NumberWeeks'] = np.around(n_weeks + n_weeks_after, decimals=2)
            single_screening['MeanEvaluation'] = inputter.get_mean_evaluation(tds)
            single_screening['StdEvaluation'] = inputter.get_std_evaluation(tds)
            single_screening['MinEvaluation'] = inputter.get_min_evaluation(tds)
            single_screening['MaxEvaluation'] = inputter.get_max_evaluation(tds)

            n_training_window = inputter.get_n_training_window(tds)
            if screening.ScreeningDate in list(tds.RatingDate):
                single_screening['NumberTraining'] = n_training_window + n_training_after
            else:
                single_screening['NumberTraining'] = n_training_window + n_training_after + 1
            single_screening['NumberTrainingWeek'] = inputter.get_n_training_week(n_weeks, n_training_window)

            training_pr_week = inputter.get_training_week(tds, start_date)
            n_weeks_with_training = inputter.get_n_weeks_with_training(tds, start_date)
            single_screening['MaxNumberTrainingWeek'] = inputter.get_n_training_week_max(training_pr_week)
            single_screening['MinNumberTrainingWeek'] = inputter.get_n_training_week_min(training_pr_week,
            n_weeks_with_training, n_weeks)
            single_screening['NumberWeeksWithTraining'] = n_weeks_with_training
            single_screening['TimeBetweenTraining'] = inputter.get_time_between_training_mean(tds)
            single_screening['NumberWeeksNoTraining'] = inputter.get_n_weeks_without_training(n_weeks,
            n_weeks_with_training)
            single_screening['NumberCancels'] = tcs.shape[0]
            single_screening['TimeBetweenCancels'] = inputter.get_mean_time_between_cancels(tcs)
            single_screening['NumberCancelsWeek'] = inputter.get_mean_cancels_week(tcs.shape[0], n_weeks)
            cancels_week = inputter.get_cancels_week(tcs)
            single_screening['MinNumberCancelsWeek'] = inputter.get_n_cancel_week_min(cancels_week)
            single_screening['MaxNumberCancelsWeek'] = inputter.get_n_training_week_max(cancels_week)

            single_screening['Ats'] = inputter.get_ats(citizen_data.ats, end_date, settings)
            single_screening['NumberAts'] = inputter.get_number_ats(citizen_data.ats, end_date)
            single_screening['LoanPeriod'] = inputter.get_avg_loan_period(citizen_data.ats, end_date)

            single_screening['Needs'] = inputter.get_needs(screening)
            single_screening['NeedsReason'] = inputter.get_needs_reason(screening)
            single_screening['Physics']= inputter.get_physics(screening)
            single_screening['PhysicsReason']= inputter.get_physics_reason(screening)

            single_screening['Ex'] = inputter.get_exercise_content(screening)
            single_screening['NumberEx'] = inputter.get_number_exercises(screening)
            ex_count = screening.ExerciseContent.count
            ex_fall = settings['fall_exercises']
            ex_threshold = settings['fall_exercise_threshold']
            single_screening['HasFallRisk'] = sum(map(ex_count, ex_fall)) > ex_threshold

            screenings = pd.concat([screenings, single_screening], axis=0, ignore_index=True)

    return screenings

def create_union_of_ids(*args: Tuple) -> List[str]:
    ids = []
    for arg in args:
        ids = list(set().union(ids, arg['CitizenId'].unique()))
    return ids

if __name__ == "__main__":
    main()
    