"""
insert_data_into_db.py
====================================
Script used to insert test data from Aalborg into our Azure db.
"""

import pandas as pd
import numpy as np
from pathlib import Path
import paths as pt
import pyodbc
from tools import preprocessor, file_reader
from datetime import date, datetime, timedelta
import random
from io import StringIO
import shutil

def main():
    # Load data
    dir = pt.RAW_DATA_DIR_TEST
    df_home_care = read_csv(dir, "Hjemmepleje.csv", encoding="iso-8859-1", skiprows=2)
    df_ats = read_csv(dir, "Hjælpemidler.csv", encoding="iso-8859-1", skiprows=2,
                      conveters={'HMI nr': str, 'Kategori ISO nummer': str})
    df_training = read_csv(dir, "Træning.csv", encoding="iso-8859-1", skiprows=2)

    df_general = df_home_care.drop_duplicates(subset=["Borger Id"])[["Borger Id",
                                                                    "Alder (aktuel)"]].reset_index(drop=True)

    df_general['Køn'] = np.random.randint(0, 2, df_general.shape[0])
    df_general['Køn'] = df_general['Køn'].apply(lambda x: "Mand" if x == 0 else "Kvinde")

    def get_first_name(gender, first_name_number):
        first_names_female = ["Jytte", "Oda", "Maren", "Karen", "Åse"]
        first_names_male = ["Karl", "Svend", "Peder", "John", "Frank"]
        if gender == "Kvinde":
            return first_names_female[first_name_number]
        else:
            return first_names_male[first_name_number]

    df_general['Fornavn'] = np.random.randint(0, 5, df_general.shape[0])
    df_general['Fornavn'] = df_general.apply(lambda x: get_first_name(x['Køn'], x['Fornavn']), axis=1)

    def get_last_name(last_name_number):
        last_names = ["Jensen", "Hansen", "Pedersen", "Nielsen", "Larsen"]
        return last_names[last_name_number]

    df_general['Efternavn'] = np.random.randint(0, 5, df_general.shape[0])
    df_general['Efternavn'] = df_general.apply(lambda x: get_last_name(x['Efternavn']), axis=1)

    def get_ssn(age, gender):
        birth_year = 2021-age
        start_date = date(birth_year, 1, 1)
        end_date = date(birth_year, 12, 31)
        time_between_dates = end_date - start_date
        days_between_dates = time_between_dates.days
        random_number_of_days = random.randrange(days_between_dates)
        random_date = start_date + timedelta(days=random_number_of_days)
        random_date = str(datetime.strftime(random_date, '%d%m%y'))
        if gender == "Kvinde":
            last_four = str(random.randint(100, 999)) + str(4)
        else:
            last_four = str(random.randint(100, 999)) + str(3)
        ssn = random_date + "-" + last_four
        return ssn

    df_general['CPR'] = df_general.apply(lambda x: get_ssn(x['Alder (aktuel)'], x['Køn']), axis=1)

    # Db
    server = "tcp:air-db-server.database.windows.net,1433"
    database = "air-db"
    username = "airadmin"
    password = "password"
    cnxn = pyodbc.connect('DRIVER={SQL Server};SERVER='+server+';DATABASE='+database+';UID='+username+';PWD='+ password)
    cursor = cnxn.cursor()

    # General data
    ids = dict()
    for _, row in df_general.iterrows():
        cursor.execute("INSERT INTO dbo.citizen (first_name,last_name,ssn,age,gender) values(?,?,?,?,?)",
                       row['Fornavn'],
                       row['Efternavn'],
                       row['CPR'],
                       row['Alder (aktuel)'],
                       row['Køn'])
        cursor.execute("SELECT @@IDENTITY AS ID;")
        ids[row['Borger Id']] = cursor.fetchone()[0]

    # Home care
    for _, row in df_home_care.iterrows():
        citizen_guid = row['Borger Id']
        citizen_id = ids[citizen_guid]
        cursor.execute("INSERT INTO dbo.home_care (citizen_id,date,type,name,time_slot) values(?,?,?,?,?)",
                citizen_id,
                row['Dato'],
                row['Paragraf'],
                row['Ydelse navn'],
                row['Bestilt tid (minutter)'].replace(",","."))

    # Training
    for _, row in df_training.iterrows():
        citizen_guid = row['Borger Id']
        citizen_id = ids[citizen_guid]
        cursor.execute("INSERT INTO dbo.training_plans (citizen_id,name,start_date,end_date,status,time_slot) values(?,?,?,?,?,?)",
                citizen_id,
                row['Ydelse navn'],
                row['Forløbsdato start'],
                row['Forløbsdato slut'],
                row['Visiteret status'],
                row['Bestilt tid (minutter)'])

    # Ats
    ats = pd.read_csv(Path.joinpath(pt.REFERENCES_DIR, 'ats.csv'), converters={'ats_id': str})
    df_ats['Kategori ISO nummer'] = df_ats['Kategori ISO nummer'].apply(lambda x: x[:6])
    df_ats['Kategori ISO navn'] = preprocessor.replace_cat_values(df_ats[['Kategori ISO nummer']], ats)
    for _, row in df_ats.iterrows():
        citizen_guid = row['Borger id']
        citizen_id = ids[citizen_guid]
        cursor.execute("INSERT INTO dbo.assistive_aids (citizen_id,name,iso,initiative_name,paragraph,lend_date) values(?,?,?,?,?,?)",
                    citizen_id,
                    row['Kategori ISO navn'],
                    row['Kategori ISO nummer'],
                    row['Indsats navn'],
                    row['Paragraf'],
                    row['Kørselsdato'])
    cnxn.commit()
    cursor.close()

def read_csv(file_path, file_name, conveters=None,
             encoding=None, skiprows=None):
    infile = StringIO()
    with open(Path.joinpath(file_path, file_name), 'r') as fd:
        shutil.copyfileobj(fd, infile)
        infile.seek(0)
        return file_reader.read_csv(infile,
                                    converters=conveters,
                                    encoding=encoding,
                                    skiprows=skiprows)

if __name__ == '__main__':
    main()
