"""
tune_alarm_rsf_wb.py
====================================
Random Survival Forest tune script for Alarm case on WanDB.
"""

from utility.config import load_config
from sksurv.ensemble import RandomSurvivalForest
from sksurv.metrics import (concordance_index_censored,
                            concordance_index_ipcw,
                            integrated_brier_score)
from sklearn.model_selection import KFold
from tools import data_loader
import paths as pt
import numpy as np
import os

os.environ["WANDB_SILENT"] = "true"
import wandb

sweep_config = {
    "method": "random", # try grid or random
    "metric": {
      "name": "c_harrell",
      "goal": "maximize"
    },
    "parameters": {
        "n_estimators": {
            "values": [50, 100, 200, 400, 600, 800, 1000]
        },
        "max_depth": {
            "values": [int(x) for x in np.linspace(1, 32, 32, endpoint=True)]
        },
        "min_samples_split": {
            "values": [float(x) for x in np.linspace(0.1, 0.9, 10, endpoint=True)]
        },
        "min_samples_leaf": {
            "values": [float(x) for x in np.linspace(0.1, 0.5, 5, endpoint=True)]
        },
        "max_features": {
            "values": [None, 'auto', 'sqrt', 'log2']
        },
    }
}

def main():
    sweep_id = wandb.sweep(sweep_config,
                           project="air-alarm-rsf")
    wandb.agent(sweep_id, train_model, count=50)

def train_model():
    config_defaults = {
        'n_estimators': [100],
        'max_depth' : [None],
        'min_samples_split': [2],
        'min_samples_leaf': [1],
        'max_features': [None],
        "seed": 0,
        "test_size": 0.3,
    }

    # Initialize a new wandb run
    wandb.init(config=config_defaults)

    # Config is a variable that holds and saves hyperparameters and inputs
    config = wandb.config

    # Load data
    target_settings = load_config(pt.CONFIGS_DIR, "alarm.yaml")
    dl = data_loader.AlarmDataLoader(pt.PROCESSED_DATA_DIR,
                                     "alarm_emb.csv",
                                     target_settings).load_data()
    X, y = dl.get_data()

    # Make model
    model = RandomSurvivalForest(n_estimators=config.n_estimators,
                                 max_depth=config.max_depth,
                                 min_samples_split=config.min_samples_split,
                                 min_samples_leaf=config.min_samples_leaf,
                                 max_features=config.max_features,
                                 random_state=0)

    # Make CV
    kf = KFold(n_splits=5, shuffle=True, random_state=0)
    c_index_harells = list()
    c_index_unos = list()
    brier_scores = list()
    for train, test in kf.split(X, y):
        model.fit(X.iloc[train], y[train])
        prediction = model.predict(X.iloc[test])

        c_harrell = concordance_index_censored(y[test]["Status"],
                                               y[test]["Days_to_alarm"],
                                               prediction)
        c_uno = concordance_index_ipcw(y[train], y[test], prediction)
        lower, upper = np.percentile(y["Days_to_alarm"], [10, 90])
        alarm_times = np.arange(lower, upper+1)
        surv_prob = np.row_stack([fn(alarm_times)
                                  for fn in model.predict_survival_function(X.iloc[test])])
        brier_score = integrated_brier_score(y[train], y[test],
                                             surv_prob, alarm_times)

        c_index_harells.append(c_harrell[0])
        c_index_unos.append(c_uno[0])
        brier_scores.append(brier_score)

    c_index_harell_mean = np.mean(c_index_harells)
    c_index_uno_mean = np.mean(c_index_unos)
    brier_score_mean = np.mean(brier_scores)

    # Log to wandb
    wandb.log({"c_harrell": c_index_harell_mean})
    wandb.log({"c_uno": c_index_uno_mean})
    wandb.log({"brier_score": brier_score_mean})

if __name__ == "__main__":
    main()
