"""
tune_complete_rf_wb.py
====================================
Random Forest tune script for Complete case on WanDB.
"""

from utility.config import load_config
from sklearn.ensemble import RandomForestClassifier
from tools import data_loader
import paths as pt
from sklearn.model_selection import cross_validate, StratifiedKFold
import numpy as np
import os

os.environ["WANDB_SILENT"] = "true"
import wandb

sweep_config = {
    "method": "random", # try grid or random
    "metric": {
      "name": "accuracy",
      "goal": "maximize"
    },
    "parameters": {
        "n_estimators": {
            "values": [1, 2, 4, 8, 16, 32, 64, 100, 200, 400]
        },
        "criterion": {
            "values": ["gini", "entropy"]
        },
        "max_depth": {
            "values": [int(x) for x in np.linspace(1, 32, 32, endpoint=True)]
        },
        "min_samples_split": {
            "values": [float(x) for x in np.linspace(0.1, 0.9, 10, endpoint=True)]
        },
        "min_samples_leaf": {
            "values": [float(x) for x in np.linspace(0.1, 0.5, 5, endpoint=True)]
        },
        "max_features": {
            "values": ['auto', 'sqrt', 'log2']
        },
        "class_weight": {
            "values": [None, "balanced"]
        },
        "bootstrap": {
            "values": [False, True]
        },
    }
}

def main():
    sweep_id = wandb.sweep(sweep_config,
                           project="air-complete-rf")
    wandb.agent(sweep_id, train_model, count=100)

def train_model():
    config_defaults = {
        'n_estimators': [100],
        'criterion' : ['gini'],
        'max_depth' : [None],
        'min_samples_split': [2],
        'min_samples_leaf': [1],
        'max_features': ['auto'],
        'class_weight': [None],
        'bootstrap': [True],
        "seed": 0,
        "test_size": 0.3,
    }

    # Initialize a new wandb run
    wandb.init(config=config_defaults)

    # Config is a variable that holds and saves hyperparameters and inputs
    config = wandb.config

    # Load data
    settings = load_config(pt.CONFIGS_DIR, "complete.yaml")
    dl = data_loader.CompleteDataLoader(pt.PROCESSED_DATA_DIR,
                                        "complete_emb.csv",
                                        settings).load_data()
    X, y = dl.get_data()

    # Make model
    model = RandomForestClassifier(n_estimators=config.n_estimators,
                                   criterion=config.criterion,
                                   max_depth=config.max_depth,
                                   min_samples_split=config.min_samples_split,
                                   min_samples_leaf=config.min_samples_leaf,
                                   max_features=config.max_features,
                                   class_weight=config.class_weight,
                                   bootstrap=config.bootstrap,
                                   random_state=0)

    # Make CV
    skf = StratifiedKFold(n_splits=5, shuffle=True, random_state=0)
    metrics = ['accuracy', 'precision', 'recall',
               'roc_auc', 'average_precision', 'f1']
    res_validate = cross_validate(model, X, y, cv=skf, scoring=metrics)

    # Evaluate performance
    accuracy = round(np.mean(res_validate[f'test_accuracy']), 3)
    precision = round(np.mean(res_validate[f'test_precision']), 3)
    recall = round(np.mean(res_validate[f'test_recall']), 3)
    roc_auc = round(np.mean(res_validate[f'test_roc_auc']), 3)
    avg_prec = round(np.mean(res_validate['test_average_precision']), 3)
    f1 = round(np.mean(res_validate['test_f1']), 3)

    # Log to wandb
    wandb.log({"accuracy": accuracy})
    wandb.log({"precision": precision})
    wandb.log({"recall": recall})
    wandb.log({"roc_auc": roc_auc})
    wandb.log({"pr_auc": avg_prec})
    wandb.log({"f1": f1})

if __name__ == "__main__":
    main()
